@@ -4,14 +4,45 @@
 //! A basic usage example can be found [here](https://github.com/hasenbanck/egui_example).
 #![warn(missing_docs)]
 
+use std::fmt::Formatter;
+use std::{borrow::Cow, num::NonZeroU32};
+
 use bytemuck::{Pod, Zeroable};
-pub use epi;
-pub use epi::egui;
-use std::borrow::Cow;
-use std::num::NonZeroU32;
 pub use wgpu;
 use wgpu::util::DeviceExt;
 
+pub use {epi, epi::egui};
+
+/// Error that the backend can return.
+#[derive(Debug)]
+pub enum BackendError {
+    /// The given `egui::TextureId` was invalid.
+    InvalidTextureId(String),
+    /// Internal implementation error.
+    Internal(String),
+}
+
+impl std::fmt::Display for BackendError {
+    fn fmt(&self, f: &mut Formatter<'_>) -> std::fmt::Result {
+        match self {
+            BackendError::InvalidTextureId(msg) => {
+                write!(f, "invalid TextureId: `{:?}`", msg)
+            }
+            BackendError::Internal(msg) => {
+                write!(f, "internal error: `{:?}`", msg)
+            }
+        }
+    }
+}
+
+impl std::error::Error for BackendError {
+    fn source(&self) -> Option<&(dyn std::error::Error + 'static)> {
+        match *self {
+            _ => None,
+        }
+    }
+}
+
 /// Enum for selecting the right buffer type.
 #[derive(Debug)]
 enum BufferType {
@@ -81,7 +112,6 @@ impl RenderPass {
         let shader = wgpu::ShaderModuleDescriptor {
             label: Some("egui_shader"),
             source: wgpu::ShaderSource::Wgsl(Cow::Borrowed(include_str!("shader/egui.wgsl"))),
-            flags: wgpu::ShaderFlags::VALIDATION,
         };
         let module = device.create_shader_module(&shader);
 
@@ -90,7 +120,7 @@ impl RenderPass {
             contents: bytemuck::cast_slice(&[UniformBuffer {
                 screen_size: [0.0, 0.0],
             }]),
-            usage: wgpu::BufferUsage::UNIFORM | wgpu::BufferUsage::COPY_DST,
+            usage: wgpu::BufferUsages::UNIFORM | wgpu::BufferUsages::COPY_DST,
         });
         let uniform_buffer = SizedBuffer {
             buffer: uniform_buffer,
@@ -102,7 +132,7 @@ impl RenderPass {
                 label: Some("egui_uniform_bind_group_layout"),
                 entries: &[wgpu::BindGroupLayoutEntry {
                     binding: 0,
-                    visibility: wgpu::ShaderStage::VERTEX,
+                    visibility: wgpu::ShaderStages::VERTEX,
                     ty: wgpu::BindingType::Buffer {
                         has_dynamic_offset: false,
                         min_binding_size: None,
@@ -131,7 +161,7 @@ impl RenderPass {
                 entries: &[
                     wgpu::BindGroupLayoutEntry {
                         binding: 0,
-                        visibility: wgpu::ShaderStage::FRAGMENT,
+                        visibility: wgpu::ShaderStages::FRAGMENT,
                         ty: wgpu::BindingType::Texture {
                             multisampled: false,
                             sample_type: wgpu::TextureSampleType::Float { filterable: true },
@@ -141,7 +171,7 @@ impl RenderPass {
                     },
                     wgpu::BindGroupLayoutEntry {
                         binding: 1,
-                        visibility: wgpu::ShaderStage::FRAGMENT,
+                        visibility: wgpu::ShaderStages::FRAGMENT,
                         ty: wgpu::BindingType::Sampler {
                             filtering: true,
                             comparison: false,
@@ -169,7 +199,7 @@ impl RenderPass {
                 module: &module,
                 buffers: &[wgpu::VertexBufferLayout {
                     array_stride: 5 * 4,
-                    step_mode: wgpu::InputStepMode::Vertex,
+                    step_mode: wgpu::VertexStepMode::Vertex,
                     // 0: vec2 position
                     // 1: vec2 texture coordinates
                     // 2: uint color
@@ -209,7 +239,7 @@ impl RenderPass {
                             operation: wgpu::BlendOperation::Add,
                         },
                     }),
-                    write_mask: wgpu::ColorWrite::ALL,
+                    write_mask: wgpu::ColorWrites::ALL,
                 }],
             }),
         });
@@ -237,7 +267,7 @@ impl RenderPass {
         paint_jobs: &[egui::paint::ClippedMesh],
         screen_descriptor: &ScreenDescriptor,
         clear_color: Option<wgpu::Color>,
-    ) {
+    ) -> Result<(), BackendError> {
         let load_operation = if let Some(color) = clear_color {
             wgpu::LoadOp::Clear(color)
         } else {
@@ -291,20 +321,21 @@ impl RenderPass {
             let height = (clip_max_y - clip_min_y).max(1);
 
             {
-                // clip scissor rectangle to target size
+                // Clip scissor rectangle to target size.
                 let x = clip_min_x.min(physical_width);
                 let y = clip_min_y.min(physical_height);
                 let width = width.min(physical_width - x);
                 let height = height.min(physical_height - y);
 
-                // skip rendering with zero-sized clip areas
+                // Skip rendering with zero-sized clip areas.
                 if width == 0 || height == 0 {
                     continue;
                 }
 
                 pass.set_scissor_rect(x, y, width, height);
             }
-            pass.set_bind_group(1, self.get_texture_bind_group(mesh.texture_id), &[]);
+            let bind_group = self.get_texture_bind_group(mesh.texture_id)?;
+            pass.set_bind_group(1, bind_group, &[]);
 
             pass.set_index_buffer(index_buffer.buffer.slice(..), wgpu::IndexFormat::Uint32);
             pass.set_vertex_buffer(0, vertex_buffer.buffer.slice(..));
@@ -312,25 +343,33 @@ impl RenderPass {
         }
 
         pass.pop_debug_group();
+
+        Ok(())
     }
 
-    fn get_texture_bind_group(&self, texture_id: egui::TextureId) -> &wgpu::BindGroup {
-        match texture_id {
-            egui::TextureId::Egui => self
-                .texture_bind_group
-                .as_ref()
-                .expect("egui texture was not set before the first draw"),
+    fn get_texture_bind_group(
+        &self,
+        texture_id: egui::TextureId,
+    ) -> Result<&wgpu::BindGroup, BackendError> {
+        let bind_group = match texture_id {
+            egui::TextureId::Egui => self.texture_bind_group.as_ref().ok_or_else(|| {
+                BackendError::Internal("egui texture was not set before the first draw".to_string())
+            })?,
             egui::TextureId::User(id) => {
                 let id = id as usize;
                 assert!(id < self.user_textures.len());
                 &(self
                     .user_textures
                     .get(id)
-                    .unwrap_or_else(|| panic!("user texture {} not found", id))
+                    .ok_or_else(|| {
+                        BackendError::Internal(format!("user texture {} not found", id))
+                    })?
                     .as_ref()
-                    .unwrap_or_else(|| panic!("user texture {} freed", id)))
+                    .ok_or_else(|| BackendError::Internal(format!("user texture {} freed", id))))?
             }
-        }
+        };
+
+        Ok(bind_group)
     }
 
     /// Updates the texture used by egui for the fonts etc. Should be called before `execute()`.
@@ -378,8 +417,8 @@ impl RenderPass {
         }
     }
 
-    // Assumes egui_texture contains srgb data.
-    // This does not match how egui::Texture is documented as of writing, but this is how it is used for user textures.
+    /// Assumes egui_texture contains srgb data.
+    /// This does not match how `egui::Texture` is documented as of writing, but this is how it is used for user textures.
     fn egui_texture_to_wgpu(
         &self,
         device: &wgpu::Device,
@@ -400,7 +439,7 @@ impl RenderPass {
             sample_count: 1,
             dimension: wgpu::TextureDimension::D2,
             format: wgpu::TextureFormat::Rgba8UnormSrgb,
-            usage: wgpu::TextureUsage::SAMPLED | wgpu::TextureUsage::COPY_DST,
+            usage: wgpu::TextureUsages::TEXTURE_BINDING | wgpu::TextureUsages::COPY_DST,
         });
 
         queue.write_texture(
@@ -408,6 +447,7 @@ impl RenderPass {
                 texture: &texture,
                 mip_level: 0,
                 origin: wgpu::Origin3d::ZERO,
+                aspect: wgpu::TextureAspect::All,
             },
             egui_texture.pixels.as_slice(),
             wgpu::ImageDataLayout {
@@ -449,9 +489,9 @@ impl RenderPass {
 
     /// Registers a `wgpu::Texture` with a `egui::TextureId`.
     ///
-    /// This enables the application to reference
-    /// the texture inside an image ui element. This effectively enables off-screen rendering inside
-    /// the egui UI. Texture must have the texture format `TextureFormat::Rgba8UnormSrgb` and
+    /// This enables the application to reference the texture inside an image ui element.
+    /// This effectively enables off-screen rendering inside the egui UI. Texture must have
+    /// the texture format `TextureFormat::Rgba8UnormSrgb` and
     /// Texture usage `TextureUsage::SAMPLED`.
     pub fn egui_texture_from_wgpu_texture(
         &mut self,
@@ -490,7 +530,64 @@ impl RenderPass {
         texture_id
     }
 
-    /// Uploads the uniform, vertex and index data used by the render pass. Should be called before `execute()`.
+    /// Registers a `wgpu::Texture` with an existing `egui::TextureId`.
+    ///
+    /// This enables applications to reuse `TextureId`s.
+    pub fn update_egui_texture_from_wgpu_texture(
+        &mut self,
+        device: &wgpu::Device,
+        texture: &wgpu::Texture,
+        texture_filter: wgpu::FilterMode,
+        id: egui::TextureId,
+    ) -> Result<(), BackendError> {
+        let id = match id {
+            egui::TextureId::User(id) => id,
+            _ => {
+                return Err(BackendError::InvalidTextureId(
+                    "ID was not of type `TextureId::User`".to_string(),
+                ));
+            }
+        };
+
+        let user_texture = self.user_textures.get_mut(id as usize).ok_or_else(|| {
+            BackendError::InvalidTextureId(format!(
+                "user texture for TextureId {} could not be found",
+                id
+            ))
+        })?;
+
+        let sampler = device.create_sampler(&wgpu::SamplerDescriptor {
+            label: Some(format!("{}_texture_sampler", self.next_user_texture_id).as_str()),
+            mag_filter: texture_filter,
+            min_filter: texture_filter,
+            ..Default::default()
+        });
+
+        // We've bound it here, so that we don't add it as a pending texture.
+        let bind_group = device.create_bind_group(&wgpu::BindGroupDescriptor {
+            label: Some(format!("{}_texture_bind_group", self.next_user_texture_id).as_str()),
+            layout: &self.texture_bind_group_layout,
+            entries: &[
+                wgpu::BindGroupEntry {
+                    binding: 0,
+                    resource: wgpu::BindingResource::TextureView(
+                        &texture.create_view(&wgpu::TextureViewDescriptor::default()),
+                    ),
+                },
+                wgpu::BindGroupEntry {
+                    binding: 1,
+                    resource: wgpu::BindingResource::Sampler(&sampler),
+                },
+            ],
+        });
+
+        *user_texture = Some(bind_group);
+
+        Ok(())
+    }
+
+    /// Uploads the uniform, vertex and index data used by the render pass.
+    /// Should be called before `execute()`.
     pub fn update_buffers(
         &mut self,
         device: &wgpu::Device,
@@ -521,7 +618,7 @@ impl RenderPass {
                 let buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
                     label: Some("egui_index_buffer"),
                     contents: data,
-                    usage: wgpu::BufferUsage::INDEX | wgpu::BufferUsage::COPY_DST,
+                    usage: wgpu::BufferUsages::INDEX | wgpu::BufferUsages::COPY_DST,
                 });
                 self.index_buffers.push(SizedBuffer {
                     buffer,
@@ -536,7 +633,7 @@ impl RenderPass {
                 let buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
                     label: Some("egui_vertex_buffer"),
                     contents: data,
-                    usage: wgpu::BufferUsage::VERTEX | wgpu::BufferUsage::COPY_DST,
+                    usage: wgpu::BufferUsages::VERTEX | wgpu::BufferUsages::COPY_DST,
                 });
 
                 self.vertex_buffers.push(SizedBuffer {
@@ -559,17 +656,17 @@ impl RenderPass {
         let (buffer, storage, name) = match buffer_type {
             BufferType::Index => (
                 &mut self.index_buffers[index],
-                wgpu::BufferUsage::INDEX,
+                wgpu::BufferUsages::INDEX,
                 "index",
             ),
             BufferType::Vertex => (
                 &mut self.vertex_buffers[index],
-                wgpu::BufferUsage::VERTEX,
+                wgpu::BufferUsages::VERTEX,
                 "vertex",
             ),
             BufferType::Uniform => (
                 &mut self.uniform_buffer,
-                wgpu::BufferUsage::UNIFORM,
+                wgpu::BufferUsages::UNIFORM,
                 "uniform",
             ),
         };
@@ -579,7 +676,7 @@ impl RenderPass {
             buffer.buffer = device.create_buffer_init(&wgpu::util::BufferInitDescriptor {
                 label: Some(format!("egui_{}_buffer", name).as_str()),
                 contents: bytemuck::cast_slice(data),
-                usage: storage | wgpu::BufferUsage::COPY_DST,
+                usage: storage | wgpu::BufferUsages::COPY_DST,
             });
         } else {
             queue.write_buffer(&buffer.buffer, 0, data);
